#include <stdio.h>
#include <stdint.h>

#define C_BEGIN 0x7E // note that this is '~' tilde
#define C_ESC   0x7D // note that this is '}' right curly brace
#define C_XOR   0x20 // special flag escaping operator

#define BUF_LEN 256
static uint8_t received[BUF_LEN];

void get_string(uint8_t buf[], uint8_t char_reg);
void put_string(uint8_t string[], uint16_t buf_len);



/* Let's send a string.*/
int main(void)
{
	uint8_t string[BUF_LEN] = "{I am a cat} Nyah~~~";

	put_string(string, BUF_LEN); // the string is sent
	printf("%s\n", (char*)received);

	return 0;
} // main()



/* This function simulates a transmission of a single byte */
void transmit(char_reg)
uint8_t char_reg;  // transmitted character
{
	get_string(received, char_reg); // the other end receives the character
} // transmit()



/* This function packages a string into a frame and sends it. */
void put_string(string, buf_len)
uint8_t string[]; // string to send
uint16_t buf_len; // length of buffer
{
	uint16_t i;
	
	transmit(C_BEGIN); // send begin flag

	for(i=0; i<buf_len; i++)
	{
		if(string[i] == C_BEGIN || string[i] == C_ESC) // need to escape 
		{
			transmit(C_ESC);              // send escape flag
			transmit(string[i] ^ C_XOR);  // send encoded byte
		}
		else // send without escaping
			transmit(string[i]); // Copy buffer to string
	} // for i
} // put_usart()



enum RxState {WAIT, ACCEPT, ESCAPE};

/* This function recevies a new character and appends the character to a
 * temporary buffer. If the message is completed (enough characters received),
 * the function copies the message from the temporary buffer to the main
 * receiving buffer.
 */
void get_string(buf, char_reg)
uint8_t buf[];      // main receiving buffer
uint8_t char_reg;   // register where new character is stored
{
	static uint8_t rx_state = WAIT;  // Start by waiting for new character.
	static uint8_t tmp_buf[BUF_LEN]; // new characters get appended here. 
	static uint16_t str_end = 0;     // index where next character goes.

	uint8_t tmp_byte = char_reg;     // copy new character from register.
	uint16_t i;


	switch(rx_state)
	{
		/* During the WAIT state, look for the header flag character. If the header
		 * flag is present, start accepting characters by going to the ACCEPT state.
		 */
		case WAIT:
			if(tmp_byte == C_BEGIN)
				rx_state = ACCEPT;
			break;

		/* During the ACCEPT state, append unescaped characters to the main buffer.
		 * If the current character is an escape character, treat the next character
		 * as an escaped character by going to the ESCAPE state.
		 *
		 * Adding functionality for recognizing additional control characters may be
		 * added here.
		 */
		case ACCEPT:
			if(tmp_byte == C_BEGIN) // header should never exist here
			{
				str_end = 0;
				rx_state = WAIT;
			}
			if(tmp_byte == C_ESC)
				rx_state = ESCAPE;	           // escape next character
			else
				tmp_buf[str_end++] = tmp_byte; // write character
			break;

		/* During the ESCAPE state, escape the current character by XOR-ing it with
		 * the special character C_XOR. Move to ACCEPT to continue.
		 */
		case ESCAPE:
			tmp_buf[str_end++] = tmp_byte ^ C_XOR;
			rx_state = ACCEPT;
			break;
	}; // switch rx_state


	/* If the temporary buffer is full after adding a new character, copy the
	 * message from the temporary buffer to a new receiving buffer.
	 *
	 * If this function is run on an interrupt, the following block must be
	 * treated as an atomic, critical section.
	 */
	if(str_end == BUF_LEN)
	{
		for(i = 0; i < BUF_LEN; i++)
			buf[i] = tmp_buf[i]; // copy completed string
		
		str_end = 0;
		rx_state = WAIT;
	} // if end of message

} // get_string
